/* $Id: PSGInstance.cc,v 1.5 2003/09/17 19:27:52 andete Exp $

lemux - LADSPA instrument plugins based on openMSX devices

(c) 2003 Joost Yervante Damad <joost@lumatec.be>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

// Emulation of a channel of the AY-3-8910
//
// code based on openMSX code, which in turn is based on
//
// code largly taken from xmame-0.37b16.1
// which is again
// based on various code snippets by Ville Hallik, Michael Cuddy,
// Tatsuyuki Satoh, Fabrice Frances, Nicola Salmoria.

#include "PSGInstance.hh"

#include <cassert>
#include <cstring>
#include <iostream>

PSGInstance::PSGInstance(unsigned long sampleRate)
:sampleRate(sampleRate),
 updateStep((FP_UNIT * sampleRate) / (CLOCK_FREQ / 8)),
 reg_env_shape_backup(-1),
 port_gate_backup(-1),
 port_amp_backup(-1),
 reg_freq_backup(-1)
{
	reset();
}

PSGInstance::~PSGInstance()
{
}

void PSGInstance::connect_port(unsigned long port, LADSPA_Data* dataLocation)
{
	switch (port)
	{
		case 0:
			port_freq = dataLocation; // audio
			break;
		case 1:
			port_amp = dataLocation;
			break;
		case 2:
			port_noise = dataLocation;
		case 3:
			port_audiotoggle = dataLocation;
			break;
		case 4:
			port_noisetoggle = dataLocation;
			break;
		case 5:
			port_env = dataLocation;
			break;
		case 6:
			port_env_freq = dataLocation;
			break;
		case 7:
			port_env_cont = dataLocation;
			break;
		case 8:
			port_env_att = dataLocation;
			break;
		case 9:
			port_env_alt = dataLocation;
			break;
		case 10:
			port_env_hold = dataLocation;
			break;
		case 11:
			port_output = dataLocation; // audio
			break;
		case 12:
			port_gate = dataLocation; // audio
			break;
		default:
			assert(false);
			break;
	}
}

void PSGInstance::activate()
{
}

void PSGInstance::run(unsigned long SampleCount)
{
	int old;
	LADSPA_Data* buf = port_output;

	// control ports //

	// noise
	int reg_noise = (int)(31*(*port_noise));
	old = periodN;
	periodN = reg_noise * updateStep;
	if (periodN == 0) periodN = updateStep;
	countN += periodN - old;
	if (countN <= 0) countN = 1;

	// volume
	int reg_vol = (int)(*port_amp);
	if (*port_amp != port_amp_backup)
	{
		vol = volTable[reg_vol];
		port_amp_backup = *port_amp;
	}

	// envelope on/off
	bool new_envelope = ((*port_env) > 0);
	if (new_envelope != envelope)
	{
		envelope = new_envelope;
		vol = envelope ? volE : volTable[reg_vol];
		validLength = 0;
	}
	
	// envelope period
	int reg_env_freq = (int)(CLOCK_FREQ/(128*(*port_env_freq)));
	old = periodE;
	periodE = reg_env_freq * (2*updateStep);
	if (periodE == 0) periodE = updateStep;
	countE += periodE - old;
	if (countE <= 0) countE = 1;

	// envelope shape
	int reg_env_shape =
		 ((*port_env_cont)>0?8:0)
		+((*port_env_att )>0?0:4)
		+((*port_env_alt )>0?2:0)
		+((*port_env_hold)>0?1:0)
		;
	
	// only recalculate if shape has changed:
	if (reg_env_shape != reg_env_shape_backup)
	{
		reg_env_shape_backup = reg_env_shape;
		attack = (reg_env_shape & 0x04) ? 0x0f : 0x00;
		if ((reg_env_shape & 0x08) == 0)
		{
			// if Continue = 0, map the shape to the equivalent one which has Continue = 1
			hold = true;
			alternate = attack;
		}
		else
		{
			hold = reg_env_shape & 0x01;
			alternate = reg_env_shape & 0x02;
		}
		countE = periodE;
		countEnv = 0x0f;
		holding = false;
		volE = volTable[countEnv ^ attack];
		if (envelope)
		{
			// std::cerr << "using envelope volume\n";
			vol = volE;  validLength = 0;
		}
	}

	// buffer calculation
	if ((*port_audiotoggle)<=0)
	{
		// audio disabled
		if (count <= SampleCount*FP_UNIT) count += SampleCount*FP_UNIT;
		output = 1;
	}
	else if (reg_vol == 0)
	{
		if (count <= SampleCount*FP_UNIT) count += SampleCount*FP_UNIT;
	}

	// for the noise channel we must not touch outputN
	// it's also not necessary since we use outn.
	if ((*port_noisetoggle)<=0)
	{
		// noise off
		if (countN <= SampleCount*FP_UNIT) countN += SampleCount*FP_UNIT;
	}

	// outn is used to randomly turn on/off the channel, in order to get noise
	bool outn = ((*port_noisetoggle)<=0);

	int i=0;
	while (SampleCount)
	{
		// audio ports //

		// freq
		int reg_freq = (int)(CLOCK_FREQ/(8*(port_freq[i])));
		if (reg_freq != reg_freq_backup)
		{
			reg_freq_backup = reg_freq;
			old = period;
			period = updateStep * reg_freq;
			if (period == 0) period = updateStep;
			count += period - old;
			if (count <= 0) count = 1;
		}

		// reset env on rising edge of gate
		if (port_gate[i] != port_gate_backup)
		{
			port_gate_backup = port_gate[i];
			if (port_gate[i] > 0 && envelope)
			{
				validLength = 0;
				countE = periodE;
				countEnv = 0x0f;
				holding = false;
				volE = volTable[countEnv ^ attack];
				vol = volE;
			}
		}
		// semiVol keeps track of how long the
		// square wave stays in the 1 position during the sample period.
		int semiVol = 0;
		int left = FP_UNIT;
		SampleCount--;
		do
		{
			int nextevent = (countN < left) ? countN : left;
			if (outn>0)
			{
				if (output) semiVol += count;
				count -= nextevent;
				// periodA is the half period of the square wave. Here, in each
				// loop I add periodA twice, so that at the end of the loop the
				// square wave is in the same status (0 or 1) it was at the start.
				// semiVolA is also incremented by periodA, since the wave has been 1
				// exactly half of the time, regardless of the initial position.
				// If we exit the loop in the middle, outputA has to be inverted
				// and semiVolA incremented only if the exit status of the square
				// wave is 1.
				while (count <= 0) {
					count += period;
					if (count > 0) {
						output ^= 1;
						if (output) semiVol += period;
						break;
					}
					count += period;
					semiVol += period;
				}
				if (output) semiVol -= count;
			} else {
				count -= nextevent;
				while (count <= 0) {
					count += period;
					if (count > 0) {
						output ^= 1;
						break;
					}
					count += period;
				}
			}
			countN -= nextevent;
			if (countN <= 0) {
				// Is noise output going to change?
				if ((random + 1) & 2)
				{	// bit0 ^ bit1
					outputN = ~outputN;
					outn = !outn;
				}
				// The Random Number Generator of the 8910 is a 17-bit shift
				// register. The input to the shift register is bit0 XOR bit2
				// (bit0 is the output).
				// The following is a fast way to compute bit 17 = bit0^bit2.
				// Instead of doing all the logic operations, we only check
				// bit 0, relying on the fact that after two shifts of the
				// register, what now is bit 2 will become bit 0, and will
				// invert, if necessary, bit 16, which previously was bit 18.
				if (random & 1) random ^= 0x28000;
				random >>= 1;
				countN += periodN;
			}

			left -= nextevent;
		} while (left > 0);

		// update envelope
		if (!holding) {
			countE -= FP_UNIT;
			if (countE <= 0) {
				do {
					countEnv--;
					countE += periodE;
				} while (countE <= 0);

				// check envelope current position
				if (countEnv < 0) {
					if (hold) {
						if (alternate) attack ^= 0x0f;
						holding = true;
						countEnv = 0;
					} else {
						// if countEnv has looped an odd number of times
						// (usually 1), invert the output.
						if (alternate && (countEnv & 0x10)) attack ^= 0x0f;
						countEnv &= 0x0f;
					}
				}
				volE = volTable[countEnv ^ attack];
				// reload volume
				if (envelope) { vol = volE; validLength = 0; }
			}
		}

		float ch = ((float)(semiVol * vol)) / FP_UNIT;
		*(buf++) = (ch/21000);
		i++;
	}
}

void PSGInstance::reset()
{
	oldEnable = 0;
	random = 1;
	output = 0;
	outputN = 0xff;
	period = periodN = periodE = 0;
	count = countN  = countE  = 0;
	setInternalVolume(21000); // value taken from Philips_NMS_8245/hardwareconfig.xml from openMSX
}

void PSGInstance::setInternalVolume(short newVolume)
{
	// calculate the volume->voltage conversion table
	// The AY-3-8910 has 16 levels, in a logarithmic scale (3dB per step)
	double out = newVolume;		// avoid clipping
	for (int i=15; i>0; i--) {
		volTable[i] = (unsigned int)(out + 0.5);	// round to nearest
		out *= 0.707945784384;			// 1/(10^(3/20)) = 1/(3dB)
		//std::cerr << volTable[i] << "\n";
	}
	volTable[0] = 0;
}

